"""The base client"""
import time
import hmac
import hashlib
import logging
from typing import List
import websockets
from orjson import loads, dumps


logger = logging.getLogger("client")


class CryptoClient:
    """Base client. Just the raw protocol features"""

    def __init__(self, api_url):
        self._next_id = 1
        self.api_url = api_url
        self.websocket = None

    @staticmethod
    def get_nonce():
        """Get the next nonce. Currently is time based"""
        return int(time.time() * 1000)

    def next_id(self):
        """Incremental id"""
        i = self._next_id
        self._next_id += 1
        return i

    def build_message(self, method, params=None, **kwargs):
        """Base message structure"""
        message = {
            "id": self.next_id(),
            "method": method,
            "nonce": self.get_nonce()
        }
        if params:
            message["params"] = params

        if kwargs:
            message.update(kwargs)
        return message

    async def send(self, message):
        """Serialize and send the message"""
        await self.websocket.send(dumps(message))

    async def next_event(self):
        """Wait for the next websocket message"""
        next_message = None
        while next_message is None:
            next_message = loads(await self.websocket.recv())
            if next_message.get("method") == "public/heartbeat":
                next_message["method"] = "public/respond-heartbeat"
                logging.info("Heartbeat")
                await self.send(next_message)
                next_message = None
        return next_message

    async def __aenter__(self):
        logger.info("Connecting to %s", self.api_url)
        self.websocket = await websockets.connect(self.api_url)
        # The api documentation recommend to wait a second in order to void rate-limit error
        time.sleep(1)
        return self

    async def __aexit__(self, exc_type, exc, atb):
        await self.websocket.close()

    async def subscribe(self, channels: List[str]):
        """Subscribe to the channels"""
        logger.info("Subscribing...")
        await self.send(self.build_message(
            method="subscribe",
            params={"channels": channels}
        ))


class AuthenticatedClient(CryptoClient):
    """Client for private user operations"""

    def __init__(self, api_key, api_secret, api_url):
        super().__init__(api_url)
        self.api_secret = api_secret.encode()
        self.api_key = api_key

    def sign_message(self, message: dict):
        """Sign message using the crypto com method"""
        message_to_sig = message["method"] + str(message["id"]) + self.api_key + str(message["nonce"])
        message["sig"] = hmac.new(self.api_secret, msg=message_to_sig.encode(), digestmod=hashlib.sha256).hexdigest()
        return message

    async def authenticate(self):
        """Perform authentication"""
        logger.info("Authenticating...")
        await self.send(self.build_message(
            method="public/auth",
            api_key=self.api_key
        ))

    async def send(self, message):
        """Serialize and send the message"""
        await self.websocket.send(dumps(self.sign_message(message)))


class MarketClient(CryptoClient):
    """Client for public operations"""

    def __init__(self, api_url="wss://stream.crypto.com/v2/market"):
        super().__init__(api_url)


class UserClient(AuthenticatedClient):
    """Client for user operations"""

    def __init__(self, api_key, api_secret, api_url="wss://stream.crypto.com/v2/user"):
        super().__init__(api_key=api_key, api_secret=api_secret, api_url=api_url)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['crypto_com']

package_data = \
{'': ['*']}

install_requires = \
['orjson>=3.5.1,<4.0.0', 'websockets>=8.1,<9.0']

setup_kwargs = {
    'name': 'crypto-com-client',
    'version': '1.0',
    'description': 'Crypto.com websocket api client',
    'long_description': 'Crypto.com websocket api client\n=================================\n[![Build Status](https://travis-ci.com/maxpowel/crypto_com_client.svg?branch=master)](https://travis-ci.com/maxpowel/crypto_com_client)\n[![Maintainability](https://api.codeclimate.com/v1/badges/9c2c51fed72ca3aeacf6/maintainability)](https://codeclimate.com/github/maxpowel/crypto_com_client/maintainability)\n[![Test Coverage](https://api.codeclimate.com/v1/badges/9c2c51fed72ca3aeacf6/test_coverage)](https://codeclimate.com/github/maxpowel/crypto_com_client/test_coverage)\n\n\nThis is a low level api client, it just connects the exchange api with your python code in the most simple way. Over\nthis library, you can build your awesome applications or high level api.\nFor more information, check the [library documentation](https://maxpowel.github.io/crypto_com_client/), the [official documentation](https://exchange-docs.crypto.com/) and the `examples` directory.\n\nFeatures\n--------\nThis library is optimized to be small, fast and secure. \n  * Fully tested: 100% code coverage\n  * Simple: It just does one thing, but it does it right\n  * Fast: Relies on asyncio so latency and memory usage is near zero (much better than threading or multiprocessing)\n  * No forced dependencies: Just `websockets` and `orjson`. No super modern cool features that you probably don\'t want\n\n\nGetting started\n---------------\nThere are two kinds of `apis`, the `user` and `market`. \nThe `user` type requires providing api credentials (access and secret key)\n\nBefore using the library, you have to install it:\n```bash\npip install crypto_com_client\n```\n\nThe most simple example, subscribing to an `orderbook`:\n\n```python\nfrom crypto_com.crypto_com import MarketClient\nimport asyncio\nimport logging\n\nlogging.basicConfig(level=logging.INFO)\n\nasync def run():\n    async with MarketClient() as client:\n        await client.subscribe(["book.CRO_USDC.10"])\n        while True:\n            event = await client.next_event()\n            print(event)\n\n\nif __name__ == "__main__":\n    loop = asyncio.get_event_loop()\n    loop.run_until_complete(run())\n```\n\nIf you want to use the `user` api first get you api `key` and `secret`.\n\n```python\nfrom crypto_com import UserClient\nimport asyncio\nimport os\nimport logging\n\nlogging.basicConfig(level=logging.INFO)\n\nasync def run():\n    async with UserClient(\n            api_key=os.environ["API_KEY"],\n            api_secret=os.environ["API_SECRET"]\n    ) as client:\n        await client.send(\n            client.build_message(\n                method="private/get-open-orders",\n                params={\n                    "instrument_name": "CRO_USDC",\n                    "page_size": 10,\n                    "page": 0\n                }\n            )\n        )\n        event = await client.next_event()\n        print(event)\n\n\nif __name__ == "__main__":\n    loop = asyncio.get_event_loop()\n    loop.run_until_complete(run())\n\n```\n\nWith these two examples you can use the whole api. Just check the API documentation to know the different methods\nand parameters.\n\nContributing\n============\nIf you have any suggestion, detect any bug or want any feature, please open an `issue` so we can discuss it.\n',
    'author': 'Alvaro Garcia',
    'author_email': 'maxpowel@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/maxpowel/crypto_com_client',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
